# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Devices on the network.
"""

from threading import RLock

from bobocep.dist.dist import BoboDistributedError

_EXC_ADDR_LEN = "address must have a length greater than 0"
_EXC_ADDR_SPACE = "address must not contain any spaces"
_EXC_PORT_RANGE = "port must be between 1 and 65535 (inclusive)"
_EXC_URN_LEN = "URN must have a length greater than 0"
_EXC_URN_SPACE = "URN must not contain any spaces"
_EXC_KEY_LEN = "ID key must have a length greater than 0"
_EXC_KEY_SPACE = "ID key must not contain any spaces"


class BoboDeviceError(BoboDistributedError):
    """
    A device error.
    """


class BoboDevice:
    """
    Contains information about a BoboCEP instance on the network.
    """

    def __init__(self,
                 addr: str,
                 port: int,
                 urn: str,
                 id_key: str):
        """
        :param addr: Device address.
        :param port: Device port.
        :param urn: Device URN.
        :param id_key: Device ID key.
        """
        super().__init__()
        self._lock: RLock = RLock()

        addr = addr.strip()

        if len(addr) == 0:
            raise BoboDeviceError(_EXC_ADDR_LEN)

        if ' ' in addr:
            raise BoboDeviceError(_EXC_ADDR_SPACE)

        if not (1 <= port <= 65535):
            raise BoboDeviceError(_EXC_URN_LEN)

        urn = urn.strip()

        if len(urn) == 0:
            raise BoboDeviceError(_EXC_URN_LEN)

        if ' ' in urn:
            raise BoboDeviceError(_EXC_URN_SPACE)

        id_key = id_key.strip()

        if len(id_key) == 0:
            raise BoboDeviceError(_EXC_KEY_LEN)

        if ' ' in id_key:
            raise BoboDeviceError(_EXC_KEY_SPACE)

        self._addr: str = addr
        self._port: int = port
        self._urn: str = urn
        self._id_key: str = id_key

    @property
    def addr(self) -> str:
        """
        :return: Device address.
        """
        with self._lock:
            return self._addr

    @addr.setter
    def addr(self, addr: str) -> None:
        """
        Set device address.

        :param addr: The new address.
        """
        with self._lock:
            addr = addr.strip()

            if len(addr) == 0:
                raise BoboDeviceError(_EXC_ADDR_LEN)

            if ' ' in addr:
                raise BoboDeviceError(_EXC_ADDR_SPACE)

            self._addr = addr

    @property
    def port(self) -> int:
        """
        :return: Device port.
        """
        return self._port

    @property
    def urn(self) -> str:
        """
        :return: Device URN.
        """
        return self._urn

    @property
    def id_key(self) -> str:
        """
        :return: Device ID key.
        """
        return self._id_key



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Distributed complex event processing.
"""

from abc import ABC, abstractmethod

from bobocep import BoboError


class BoboDistributedError(BoboError):
    """
    A distributed error.
    """


class BoboDistributedSystemError(BoboDistributedError):
    """
    A distributed system error.
    """


class BoboDistributedTimeoutError(BoboDistributedError):
    """
    A distributed timeout error.
    """


class BoboDistributedJSONError(BoboDistributedError):
    """
    A distributed JSON error.
    """


class BoboDistributedJSONEncodeError(BoboDistributedJSONError):
    """
    A distributed JSON encode error.
    """


class BoboDistributedJSONDecodeError(BoboDistributedJSONError):
    """
    A distributed JSON decode error.
    """


class BoboDistributed(ABC):
    """
    Distributed `BoboCEP`.
    """

    @abstractmethod
    def run(self) -> None:
        """
        Runs distributed.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Generator imports.
"""

from bobocep.cep.gen.event import BoboGenEvent, BoboGenEventNone, \
    BoboGenEventTime
from bobocep.cep.gen.event_id import BoboGenEventID, BoboGenEventIDUnique
from bobocep.cep.gen.timestamp import BoboGenTimestamp, BoboGenTimestampEpoch



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Event imports.
"""

from bobocep.cep.event.action import BoboEventAction
from bobocep.cep.event.complex import BoboEventComplex
from bobocep.cep.event.event import BoboEvent, BoboEventError
from bobocep.cep.event.factory import BoboEventFactory, BoboEventFactoryError
from bobocep.cep.event.history import BoboHistory
from bobocep.cep.event.simple import BoboEventSimple



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Multi actions.
"""
from abc import ABC
from typing import Tuple, Any, List

from bobocep.cep.action import BoboAction, BoboActionError
from bobocep.cep.event import BoboEventComplex

_ACTIONS_MIN: int = 1


class BoboActionMulti(BoboAction, ABC):
    """
    An abstract multi action.
    """

    def __init__(self, name: str, *args, **kwargs):
        """
        :param name: The action name.
        :param args: Action arguments.
        :param kwargs: Action keyword arguments.
        """
        super().__init__(name=name, args=args, kwargs=kwargs)


class BoboActionMultiSequential(BoboActionMulti):
    """
    An abstract sequential multi action.
    """

    def __init__(self,
                 name: str,
                 actions: List[BoboAction],
                 stop_on_fail: bool,
                 *args,
                 **kwargs):
        """
        :param name: The action name.
        :param actions: The list of actions to execute.
        :param stop_on_fail: If True, the multi-action stops processing its
            action list if its current action fails. If False, it continues
            to process its remaining actions. Note: failure of any action in
            its list will cause the multi-action's success to be False.
        :param args: Action arguments.
        :param kwargs: Action keyword arguments.
        """
        super().__init__(name=name, args=args, kwargs=kwargs)

        if len(actions) < 1:
            raise BoboActionError(
                f"multi sequential action {name} "
                f"must contain at least {_ACTIONS_MIN} action")

        self._actions: List[BoboAction] = actions
        self._stop_on_fail: bool = stop_on_fail

    def execute(self, event: BoboEventComplex) \
            -> Tuple[bool, List[Tuple[bool, Any]]]:
        """
        :param event: The complex event that triggered action.
        :return: Whether the action execution was successful, and
            any additional data.
        """
        success = True
        data: List[Tuple[bool, Any]] = []

        for action in self._actions:
            output: Tuple[bool, Any] = action.execute(event)
            data.append(output)

            # If action was unsuccessful
            if not output[0]:
                success = False

                if self._stop_on_fail:
                    break

        return success, data



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Builders to assist in generating a pattern.
"""
import typing
from typing import List, Union, Callable

from bobocep.cep.phenom.pattern.pattern import BoboPatternBlock, \
    BoboPattern, BoboPatternError
from bobocep.cep.phenom.pattern.predicate import BoboPredicate, \
    BoboPredicateCall

_EXC_NAME_LEN = "name must have a length greater than 0"


# Decorator "@typing.no_type_check" is used to suppress a mypy issue regarding
# the use of "Union[BoboPredicate, Callable]" for predicate parameters.
# This may have the unintended side effect of suppressing actual typing issues
# present in the methods.


class BoboPatternBuilderError(BoboPatternError):
    """
    A pattern builder error.
    """


class BoboPatternBuilder:
    """
    A pattern builder.
    """

    def __init__(self, name: str):
        """
        :param name: Pattern name.
        """

        if len(name) == 0:
            raise BoboPatternBuilderError(_EXC_NAME_LEN)

        self._name: str = name
        self._blocks: List[BoboPatternBlock] = []
        self._preconditions: List[BoboPredicate] = []
        self._haltconditions: List[BoboPredicate] = []

    def generate(self) -> BoboPattern:
        """
        Generates a BoboPattern instance with the configuration specified
        in the builder.

        :return: A BoboPattern instance.
        """

        return BoboPattern(
            name=self._name,
            blocks=self._blocks,
            preconditions=self._preconditions,
            haltconditions=self._haltconditions
        )

    @typing.no_type_check
    def next(self,
             predicate: Union[BoboPredicate, Callable],
             group: str = "",
             times: int = 1,
             loop: bool = False) -> 'BoboPatternBuilder':
        """
        Adds a block with strict contiguity.

        :param predicate: Block predicate. If a Callable is provided, it will
            be wrapped in a BoboPredicateCall instance.
        :param group: Block group.
        :param times: Number of times to add this block to the pattern,
            in sequence.
        :param loop: If `True`, the block loops back onto itself (making it
            non-deterministic). If `False`, the block remains deterministic.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=[predicate],
                group=group,
                strict=True,
                loop=loop,
                negated=False,
                optional=False))
        return self

    @typing.no_type_check
    def not_next(self,
                 predicate: Union[BoboPredicate, Callable],
                 group: str = "",
                 times: int = 1) -> 'BoboPatternBuilder':
        """
        Adds a negated block with strict contiguity.

        :param predicate: Block predicate. If a Callable is provided, it will
            be wrapped in a BoboPredicateCall instance.
        :param group: Block group.
        :param times: Number of times to add this block to the pattern,
            in sequence.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=[predicate],
                group=group,
                strict=True,
                loop=False,
                negated=True,
                optional=False))
        return self

    @typing.no_type_check
    def followed_by(
            self,
            predicate: Union[BoboPredicate, Callable],
            group: str = "",
            times: int = 1,
            loop: bool = False,
            optional: bool = False) -> 'BoboPatternBuilder':
        """
        Adds a block with relaxed contiguity.

        :param predicate: Block predicate. If a Callable is provided, it will
            be wrapped in a BoboPredicateCall instance.
        :param group: Block group.
        :param times: Number of times to add this block to the pattern,
            in sequence.
        :param loop: If `True`, the block loops back onto itself.
        :param optional: If `True`, the block is optional.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=[predicate],
                group=group,
                strict=False,
                loop=loop,
                negated=False,
                optional=optional))
        return self

    @typing.no_type_check
    def not_followed_by(
            self,
            predicate: Union[BoboPredicate, Callable],
            group: str = "",
            times: int = 1) -> 'BoboPatternBuilder':
        """
        Adds a negated block with relaxed contiguity.

        :param predicate: Block predicate. If a Callable is provided, it will
            be wrapped in a BoboPredicateCall instance.
        :param group: Block group.
        :param times: Number of times to add this block to the pattern,
            in sequence.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=[predicate],
                group=group,
                strict=False,
                loop=False,
                negated=True,
                optional=False))
        return self

    @typing.no_type_check
    def followed_by_any(
            self,
            predicates: List[Union[BoboPredicate, Callable]],
            group: str = "",
            times: int = 1,
            loop: bool = False,
            optional: bool = False) -> 'BoboPatternBuilder':
        """
        Adds multiple blocks with non-deterministic relaxed contiguity.

        :param predicates: Predicates, one per block. Any Callable types in
            the list will be wrapped in their own BoboPredicateCall instance.
        :param group: Group name for all blocks.
        :param times: Number of times to add these blocks to the pattern,
            in sequence.
        :param loop: If `True`, the blocks loop back onto themselves.
        :param optional: If `True`, the blocks are optional.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        for i in range(len(predicates)):
            if isinstance(predicates[i], Callable):
                predicates[i] = BoboPredicateCall(
                    call=predicates[i])

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=predicates,
                group=group,
                strict=False,
                loop=loop,
                negated=False,
                optional=optional))
        return self

    @typing.no_type_check
    def not_followed_by_any(
            self,
            predicates: List[Union[BoboPredicate, Callable]],
            group: str = "",
            times: int = 1) -> 'BoboPatternBuilder':
        """
        Adds multiple negated blocks with non-deterministic relaxed contiguity.

        :param predicates: Predicates, one per block. Any Callable types in
            the list will be wrapped in their own BoboPredicateCall instance.
        :param group: Group name for all blocks.
        :param times: Number of times to add these blocks to the pattern,
            in sequence.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        for i in range(len(predicates)):
            if isinstance(predicates[i], Callable):
                predicates[i] = BoboPredicateCall(
                    call=predicates[i])

        for _ in range(max(times, 1)):
            self._blocks.append(BoboPatternBlock(
                predicates=predicates,
                group=group,
                strict=False,
                loop=False,
                negated=True,
                optional=False))
        return self

    @typing.no_type_check
    def precondition(
            self, predicate: Union[BoboPredicate, Callable]) \
            -> 'BoboPatternBuilder':
        """
        Adds a precondition.

        :param predicate: The precondition predicate.
            If a Callable is provided, it will be wrapped in a
            BoboPredicateCall instance.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        self._preconditions.append(predicate)
        return self

    @typing.no_type_check
    def haltcondition(
            self, predicate: Union[BoboPredicate, Callable]) \
            -> 'BoboPatternBuilder':
        """
        Adds a haltcondition.

        :param predicate: The haltcondition predicate.
            If a Callable is provided, it will be wrapped in a
            BoboPredicateCall instance.

        :return: The BoboPatternBuilder instance that made the function call.
        """

        if isinstance(predicate, Callable):
            predicate = BoboPredicateCall(call=predicate)

        self._haltconditions.append(predicate)
        return self



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Event factory.
"""

from json import loads

from bobocep.cep.event.event import BoboEventError, BoboEvent


class BoboEventFactoryError(BoboEventError):
    """
    An event factory error.
    """


class BoboEventFactory:
    """
    A BoboEvent factory that generates instances from JSON representations
    of events.
    """

    @staticmethod
    def from_json_str(j: str) -> BoboEvent:
        """
        :param j: A JSON `str` representation of the event.
        :return: A new instance of the event type.

        :raises BoboEventFactoryError: If `EVENT_TYPE` key is missing
            from JSON.
        :raises BoboEventFactoryError: If `EVENT_TYPE` value is an unknown
            event type.
        """
        from bobocep.cep.event.action import BoboEventAction
        from bobocep.cep.event.complex import BoboEventComplex
        from bobocep.cep.event.simple import BoboEventSimple

        d: dict = loads(j)

        if BoboEvent.EVENT_TYPE not in d:
            raise BoboEventFactoryError(
                "Missing key '{}'.".format(BoboEvent.EVENT_TYPE))

        if d[BoboEvent.EVENT_TYPE] == BoboEventSimple.TYPE_SIMPLE:
            return BoboEventSimple.from_json_dict(d)

        if d[BoboEvent.EVENT_TYPE] == BoboEventComplex.TYPE_COMPLEX:
            return BoboEventComplex.from_json_dict(d)

        if d[BoboEvent.EVENT_TYPE] == BoboEventAction.TYPE_ACTION:
            return BoboEventAction.from_json_dict(d)

        raise BoboEventFactoryError(
            "Unknown event type '{}'.".format(d[BoboEvent.EVENT_TYPE]))



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
A serializable representation of a run.
"""

from json import loads, dumps

from bobocep import BoboError
from bobocep.bobocep import BoboJSONable
from bobocep.cep.event import BoboHistory

_EXC_RUN_ID_LEN = "run ID must have a length greater than 0"
_EXC_PHENOM_LEN = "phenomenon name must have a length greater than 0"
_EXC_INDEX = "block index must be greater than 1"
_EXC_HISTORY_LEN = "history must have at least 1 event"


class BoboRunSerialError(BoboError):
    """
    A run serial error.
    """


class BoboRunSerial(BoboJSONable):
    """
    Represents the state of a run and is designed to be serializable.
    """

    RUN_ID = "run_id"
    PHENOMENON_NAME = "phenomenon_name"
    PATTERN_NAME = "pattern_name"
    BLOCK_INDEX = "block_index"
    HISTORY = "history"

    def __init__(self,
                 run_id: str,
                 phenomenon_name: str,
                 pattern_name: str,
                 block_index: int,
                 history: BoboHistory):
        """
        :param run_id: An ID for the run.
        :param phenomenon_name: A phenomenon name associated with the run.
        :param pattern_name: A pattern name associated with the run.
        :param block_index: An index which indicates where in the pattern
            to start the run.
        :param history: A history of events for the run.

        :raises BoboRunError: Run ID length is equal to 0.
        :raises BoboRunError: Process name length is equal to 0.
        :raises BoboRunError: Block index is less than 1.
        :raises BoboRunError: History does not have enough events in it
            to cover all blocks up to the block index.
        """
        super().__init__()

        if len(run_id) == 0:
            raise BoboRunSerialError(_EXC_RUN_ID_LEN)

        if len(phenomenon_name) == 0:
            raise BoboRunSerialError(_EXC_PHENOM_LEN)

        if block_index < 1:
            raise BoboRunSerialError(_EXC_INDEX)

        if history.size() < 1:
            raise BoboRunSerialError(_EXC_HISTORY_LEN)

        self._run_id: str = run_id
        self._phenomenon_name: str = phenomenon_name
        self._pattern_name: str = pattern_name
        self._block_index: int = block_index
        self._history: BoboHistory = history

    @property
    def run_id(self) -> str:
        """
        :return: The run ID.
        """
        return self._run_id

    @property
    def phenomenon_name(self) -> str:
        """
        :return: The phenomenon name associated with the run.
        """
        return self._phenomenon_name

    @property
    def pattern_name(self) -> str:
        """
        :return: The pattern name associated with the run.
        """
        return self._pattern_name

    @property
    def block_index(self) -> int:
        """
        :return: The current block index of the run.
        """
        return self._block_index

    @property
    def history(self) -> BoboHistory:
        """
        :return: The run history.
        """
        return self._history

    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of the run.
        """
        return {
            self.RUN_ID: self.run_id,
            self.PHENOMENON_NAME: self.phenomenon_name,
            self.PATTERN_NAME: self.pattern_name,
            self.BLOCK_INDEX: self.block_index,
            self.HISTORY: self.history
        }

    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of the run.
        """
        return dumps(self.to_json_dict(), default=lambda o: o.to_json_str())

    @staticmethod
    def from_json_str(j: str) -> 'BoboRunSerial':
        """
        :param j: A JSON `str` representation of the event.

        :return: A new instance of the run serial.
        """
        return BoboRunSerial.from_json_dict(loads(j))

    @staticmethod
    def from_json_dict(d: dict) -> 'BoboRunSerial':
        """
        :param d: A JSON `dict` representation of the event.

        :return: A new instance of the run serial.
        """
        return BoboRunSerial(
            run_id=d[BoboRunSerial.RUN_ID],
            phenomenon_name=d[BoboRunSerial.PHENOMENON_NAME],
            pattern_name=d[BoboRunSerial.PATTERN_NAME],
            block_index=d[BoboRunSerial.BLOCK_INDEX],
            history=BoboHistory.from_json_str(d[BoboRunSerial.HISTORY])
        )

    def __str__(self) -> str:
        """
        :return: A JSON `str` representation of the run.
        """
        return self.to_json_str()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
AES encryption for distributed traffic.
"""
from threading import RLock

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from bobocep.dist.crypto.crypto import BoboDistributedCrypto, \
    BoboDistributedCryptoError

_UTF_8: str = "UTF-8"
_PAD_MODULO: int = 16
_END_BYTES: bytes = "BOBO".encode(_UTF_8)
_LEN_END_BYTES: int = len(_END_BYTES)
_PAD_CHAR: str = '\0'

_BYTES_AES_128: int = 16
_BYTES_AES_192: int = 24
_BYTES_AES_256: int = 32


class BoboDistributedCryptoAES(BoboDistributedCrypto):
    """
    AES encryption in GCM mode.
    Data are encrypted using either AES-128, AES-192, or AES-256 encryption.
    """

    def __init__(self,
                 aes_key: str,
                 nonce_length: int = 16,
                 mac_length: int = 16):
        """
        :param aes_key: The AES key to use for encryption.
        :param nonce_length: The nonce length.
        :param mac_length: The MAC length.
        """
        super().__init__()
        self._lock: RLock = RLock()

        # Check AES key
        num_bytes_aes_key = len(aes_key)
        if (num_bytes_aes_key != _BYTES_AES_128 and
                num_bytes_aes_key != _BYTES_AES_192 and
                num_bytes_aes_key != _BYTES_AES_256):
            raise BoboDistributedCryptoError(
                "AES key must be 16, 24, or 32 bytes long "
                "(respectively for AES-128, AES-192, or AES-256): "
                "'{}' has {} bytes.".format(aes_key, len(aes_key)))

        self._aes_key: bytes = aes_key.encode(_UTF_8)
        self._nonce_length: int = nonce_length
        self._mac_length: int = mac_length

        self._msg_min_length: int = (
                _PAD_MODULO +
                nonce_length +
                mac_length +
                _LEN_END_BYTES
        )

    def end_bytes(self) -> bytes:
        """
        :return: Bytes used to signify the end of every encrypted payload.
        """
        return _END_BYTES

    def min_length(self) -> int:
        """
        :return: The minimum possible length of an encrypted payload,
            in number of bytes.
        """
        return self._msg_min_length

    def encrypt(self, msg_str: str) -> bytes:
        """
        :param msg_str: Wraps JSON string in other data for transit.
        :return: The bytes to transmit.
        """
        nonce = get_random_bytes(self._nonce_length)
        cipher = AES.new(self._aes_key, AES.MODE_GCM,
                         nonce=nonce, mac_len=self._mac_length)

        # Pad message, if necessary
        len_msg = len(msg_str)
        if len_msg % _PAD_MODULO != 0:
            msg_str = msg_str + \
                      (_PAD_CHAR * (_PAD_MODULO - len_msg % _PAD_MODULO))

        ciphertext, mac = cipher.encrypt_and_digest(  # type: ignore
            msg_str.encode(_UTF_8))

        # Append nonce, mac, and end bytes to ciphertext
        ciphertext_bytes = bytearray(ciphertext)
        ciphertext_bytes.extend(nonce)
        ciphertext_bytes.extend(mac)
        ciphertext_bytes.extend(_END_BYTES)

        return ciphertext_bytes

    def decrypt(self, msg_bytes: bytes) -> str:
        """
        :param msg_bytes: Incoming bytes.
        :return: Incoming JSON string with other data from transit.
        """

        # [CIPHERTEXT][NONCE][MAC][END_BYTES]
        ciphertext = msg_bytes[:-(
                self._nonce_length + self._mac_length + _LEN_END_BYTES)]
        nonce = msg_bytes[
                -(self._nonce_length + self._mac_length + _LEN_END_BYTES):
                -(self._mac_length + _LEN_END_BYTES)]
        mac = msg_bytes[
              -(self._mac_length + _LEN_END_BYTES):
              -_LEN_END_BYTES]

        cipher = AES.new(self._aes_key, AES.MODE_GCM, nonce=nonce)
        plaintext = cipher.decrypt_and_verify(  # type: ignore
            ciphertext, mac).decode(_UTF_8).rstrip(_PAD_CHAR)

        return plaintext



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Generates BoboEvent timestamps.
"""

from abc import ABC, abstractmethod
from time import time_ns


class BoboGenTimestamp(ABC):
    """
    A timestamp generator.
    """

    @abstractmethod
    def generate(self) -> int:
        """
        :return: Generated timestamp.
        """


class BoboGenTimestampEpoch(BoboGenTimestamp):
    """
    A timestamp generator that returns the current time, in milliseconds, since
    the Epoch.
    """

    def generate(self) -> int:
        """
        :return: The current time, in milliseconds, since the Epoch.
        """
        return time_ns() // 1000000



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Action imports.
"""

from bobocep.cep.action.action import BoboAction, BoboActionError
from bobocep.cep.action.handler import BoboActionHandler, \
    BoboActionHandlerBlocking, BoboActionHandlerMultithreading, \
    BoboActionHandlerMultiprocessing, BoboActionHandlerError



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Setup imports.
"""

from bobocep.setup.setup import BoboSetupError
from bobocep.setup.simple import BoboSetupSimple, BoboSetupSimpleDistributed



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Tests that ensure complex event generation occurs when expected.
The tests contain typical patterns that may be built by the user
with the pattern builder.

This is not an exhaustive list due to the sheer number of possible
combinations. However, the tests are representative of some of the
common patterns that are likely to frequently occur.
"""

from threading import RLock
from typing import Tuple, Any, List

from bobocep.cep.action import BoboAction, BoboActionHandlerBlocking
from bobocep.cep.event import BoboEventComplex
from bobocep.cep.phenom import BoboPattern, BoboPhenomenon
from bobocep.cep.phenom.pattern.builder import BoboPatternBuilder
from bobocep.setup import BoboSetupSimple


class BoboActionCounter(BoboAction):
    """
    An action that counts how many times it has been executed.
    """

    def __init__(self, name: str):
        """
        :param name: The name of the action.
        """
        super().__init__(name)
        self._lock: RLock = RLock()
        self._counter: int = 0

    def execute(self, event: BoboEventComplex) -> Tuple[bool, Any]:
        """
        Increments the counter value.

        :param event: The complex event.

        :return: True, and the new counter value
        """
        with self._lock:
            self._counter += 1

            return True, None

    @property
    def counter(self) -> int:
        """
        Get counter.
        """
        with self._lock:
            return self._counter


def _setup(patterns: List[BoboPattern]):
    action = BoboActionCounter(name="action")

    phenom = BoboPhenomenon(
        name="phenom",
        patterns=patterns,
        action=action)

    engine = BoboSetupSimple(
        phenomena=[phenom],
        handler=BoboActionHandlerBlocking()
    ).generate()

    return engine, action


class TestValid:

    def test_fb_fb_fb_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_nfb_fb_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .not_followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 6, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_nfb_fb_data_not_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .not_followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 6, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fba_fb_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by_any([
                lambda e, h: int(e.data) == 21,
                lambda e, h: int(e.data) == 22,
                lambda e, h: int(e.data) == 23
            ]) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 22, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fba_fb_data_not_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by_any([
                lambda e, h: int(e.data) == 21,
                lambda e, h: int(e.data) == 22,
                lambda e, h: int(e.data) == 23
            ]) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 6, 23, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_nfba_fb_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .not_followed_by_any([
                lambda e, h: int(e.data) == 21,
                lambda e, h: int(e.data) == 22,
                lambda e, h: int(e.data) == 23
            ]) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_nfba_fb_data_not_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .not_followed_by_any([
                lambda e, h: int(e.data) == 21,
                lambda e, h: int(e.data) == 22,
                lambda e, h: int(e.data) == 23
            ]) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 21, 22, 23, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_n_n_n_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .next(lambda e, h: int(e.data) == 1) \
            .next(lambda e, h: int(e.data) == 2) \
            .next(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_n_nn_n_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .next(lambda e, h: int(e.data) == 1) \
            .not_next(lambda e, h: int(e.data) == 2) \
            .next(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 6, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_pre_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .precondition(lambda e, h: int(e.data) > 0) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_halt_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .haltcondition(lambda e, h: int(e.data) == 10) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_times_data_exact(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, times=3) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 2, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_loop_minimum(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, loop=True) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_loop_once(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, loop=True) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_loop_many(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, loop=True) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_optional_included(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, optional=True) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1

    def test_fb_fb_fb_optional_not_included(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2, optional=True) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 1


class TestInvalid:

    def test_fb_fb_fb_pre(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .precondition(lambda e, h: int(e.data) > 0) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 0, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 0

    def test_fb_fb_fb_halt(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .followed_by(lambda e, h: int(e.data) == 1) \
            .followed_by(lambda e, h: int(e.data) == 2) \
            .followed_by(lambda e, h: int(e.data) == 3) \
            .haltcondition(lambda e, h: int(e.data) == 10) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 10, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 0

    def test_n_n_n(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .next(lambda e, h: int(e.data) == 1) \
            .next(lambda e, h: int(e.data) == 2) \
            .next(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 6, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 0

    def test_n_nn_n(self):
        pattern_1: BoboPattern = BoboPatternBuilder("pattern_1") \
            .next(lambda e, h: int(e.data) == 1) \
            .not_next(lambda e, h: int(e.data) == 2) \
            .next(lambda e, h: int(e.data) == 3) \
            .generate()

        engine, action = _setup([pattern_1])

        for i in [1, 2, 3]:
            engine.receiver.add_data(i)
            engine.update()

        assert len(engine.decider.all_runs()) == 0
        assert action.counter == 0



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
A run.
"""

from threading import RLock
from typing import Dict, Tuple, List

from bobocep import BoboError
from bobocep.cep.engine.decider.runserial import BoboRunSerial
from bobocep.cep.event import BoboHistory, BoboEvent
from bobocep.cep.phenom.pattern.pattern import BoboPattern, \
    BoboPatternBlock, BoboPredicate

_EXC_RUN_ID_LEN = "run ID must have a length greater than 0"
_EXC_PHENOM_LEN = "phenomenon name must have a length greater than 0"
_EXC_INDEX = "block index must be greater than 1"
_EXC_HISTORY_LEN = "history must have at least 1 event"


class BoboRunError(BoboError):
    """
    A run error.
    """


class BoboRun:
    """
    A run that tracks the progress of a partially completed complex event.
    """

    def __init__(self,
                 run_id: str,
                 phenomenon_name: str,
                 pattern: BoboPattern,
                 block_index: int,
                 history: BoboHistory):
        """
        :param run_id: An ID for the run.
        :param phenomenon_name: A phenomenon name associated with the run.
        :param pattern: A pattern associated with the run.
        :param block_index: An index which indicates where in the pattern
            to start the run.
        :param history: A history of events for the run.

        :raises BoboRunError: Run ID length is equal to 0.
        :raises BoboRunError: Process name length is equal to 0.
        :raises BoboRunError: Block index is less than 1.
        :raises BoboRunError: History does not have enough events in it
            to cover all blocks up to the block index.
        """
        super().__init__()
        self._lock: RLock = RLock()

        if len(run_id) == 0:
            raise BoboRunError(_EXC_RUN_ID_LEN)

        if len(phenomenon_name) == 0:
            raise BoboRunError(_EXC_PHENOM_LEN)

        if block_index < 1:
            raise BoboRunError(_EXC_INDEX)

        if history.size() < 1:
            raise BoboRunError(_EXC_HISTORY_LEN)

        self._run_id: str = run_id
        self._phenomenon_name: str = phenomenon_name
        self._pattern: BoboPattern = pattern
        self._block_index: int = block_index
        self._history: BoboHistory = history
        self._halted: bool = self.is_complete()

    @property
    def run_id(self) -> str:
        """
        :return: The run ID.
        """
        return self._run_id

    @property
    def phenomenon_name(self) -> str:
        """
        :return: The phenomenon name associated with the run.
        """
        return self._phenomenon_name

    @property
    def pattern(self) -> BoboPattern:
        """
        :return: The pattern associated with the run.
        """
        return self._pattern

    @property
    def block_index(self) -> int:
        """
        :return: The current block index of the run.
        """
        with self._lock:
            return self._block_index

    def set_block(self, block_index: int, history: BoboHistory) -> None:
        """
        Updates the run's block to a new index and history.

        :param block_index: The new block index.
        :param history: The new history.

        :raises: BoboRunError: New block index is less than 1.
        :raises: BoboRunError: New history does not have enough events in it
            to cover all blocks up to the new block index.
        """
        with self._lock:
            if block_index < 1:
                raise BoboRunError(_EXC_INDEX.format)

            self._block_index = block_index
            self._history = history

    def history(self) -> BoboHistory:
        """
        :return: The run history.
        """
        with self._lock:
            return self._history

    def is_complete(self) -> bool:
        """
        :return: True if run is completed; False otherwise.
        """
        with self._lock:
            return self._block_index > len(self.pattern.blocks) - 1

    def is_halted(self) -> bool:
        """
        :return: True if run has halted; False otherwise.
        """
        with self._lock:
            return self._halted

    def serialize(self) -> BoboRunSerial:
        """
        :return: A serializable representation of the run.
        """

        with self._lock:
            return BoboRunSerial(
                run_id=self.run_id,
                phenomenon_name=self.phenomenon_name,
                pattern_name=self.pattern.name,
                block_index=self._block_index,
                history=self._history)

    def halt(self) -> None:
        """
        Halts the run.
        """
        with self._lock:
            self._halted = True

    def process(self, event: BoboEvent) -> bool:
        """
        :param event: An event for the run to process.
        :return: `True` if the event caused a state change in the run;
            `False` otherwise.
        """
        # True if state change occurred; False otherwise.
        # E.g. accepted event, changed block, completed, halted.
        with self._lock:
            # Do not process if the run has already finished
            if self._halted:
                return False

            # Halt if run does not match against all preconditions
            if len(self.pattern.preconditions) > 0:
                if not all([precon.evaluate(event, self._history)
                            for precon in self.pattern.preconditions]):
                    self._halted = True
                    return True

            # Halt if run matches against any haltconditions
            if len(self.pattern.haltconditions) > 0:
                if any([haltcon.evaluate(event, self._history)
                        for haltcon in self.pattern.haltconditions]):
                    self._halted = True
                    return True

            temp_index = self._block_index
            block: BoboPatternBlock = self.pattern.blocks[temp_index]

            if block.loop:
                return self._process_loop(event, block, temp_index)
            else:
                return self._process_not_loop(event, block, temp_index)

    def _process_loop(self,
                      event: BoboEvent,
                      block: BoboPatternBlock,
                      temp_index: int) -> bool:
        """
        Process a block that loops.

        :param event: An event.
        :param block: The block to process.
        :param temp_index: Temporary index used during processing.

        :return: `True` if halted; `False` otherwise.
        """
        match = self._is_match(event, block.predicates)

        if match:
            self._add_event(event, block)
            return True
        else:
            # Looping block can be neither negated nor optional.
            if block.strict:
                self._halted = True
                return True
            else:
                # Looping block cannot be the final state.
                temp_index += 1
                block = self.pattern.blocks[temp_index]
                if block.loop:
                    return self._process_loop(event, block, temp_index)
                else:
                    return self._process_not_loop(event, block, temp_index)

    def _process_not_loop(self,
                          event: BoboEvent,
                          block: BoboPatternBlock,
                          temp_index: int) -> bool:
        """
        Process a block that does not loop.

        :param event: An event.
        :param block: The block to process.
        :param temp_index: Temporary index used during processing.

        :return: `True` if halted; `False` otherwise.
        """
        # Non-looping block cannot be both negated and optional.
        match = self._is_match(event, block.predicates)

        if block.negated:
            if match:
                # Negated predicate happened: failure.
                if block.strict:
                    self._halted = True
                    return True
                return False
            else:
                # Negated predicate did not happen: success.
                self._move_forward(event, block, temp_index)
                return True

        elif block.optional:
            # Strict block cannot be optional.
            if match:
                # Optional block consumes event.
                self._move_forward(event, block, temp_index)
                return True
            else:
                # Try event against next block.
                temp_index += 1
                block = self.pattern.blocks[temp_index]
                if block.loop:
                    return self._process_loop(event, block, temp_index)
                else:
                    return self._process_not_loop(event, block, temp_index)
        else:
            if match:
                self._move_forward(event, block, temp_index)
                return True
            else:
                if block.strict:
                    self._halted = True
                    return True
                return False

    def _is_match(self,
                  event: BoboEvent,
                  predicates: Tuple[BoboPredicate, ...]) -> bool:
        """
        :param event: An event.
        :param predicates: Predicates to match against.

        :return: `True` if the event matches any predicate;
            `False` otherwise.
        """
        return any(predicate.evaluate(event, self._history)
                   for predicate in predicates)

    def _add_event(self, event: BoboEvent, block: BoboPatternBlock) -> None:
        """
        :param event: Event to add to history.
        :param block: Block to determine which group to add event to
            in the history.
        """
        newevents: Dict[str, List[BoboEvent]] = self._history.events

        if block.group not in newevents:
            newevents[block.group] = []

        newevents[block.group].append(event)
        self._history = BoboHistory(events=newevents)

    def _move_forward(self,
                      event: BoboEvent,
                      block: BoboPatternBlock,
                      temp_index: int) -> None:
        """
        Move run forward to a new block.

        :param event: An event.
        :param block: A block.
        :param temp_index: Temporary index used during processing.
        """
        self._add_event(event, block)
        self._block_index = temp_index + 1
        self._halted = self.is_complete()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Forwarder publish-subscribe classes.
"""

from abc import ABC, abstractmethod

from bobocep.cep.event import BoboEventAction


class BoboForwarderSubscriber(ABC):
    """
    A forwarder subscriber interface.
    """

    @abstractmethod
    def on_forwarder_update(self, event: BoboEventAction) -> None:
        """
        :param event: Action event generated by Forwarder.
        """


class BoboForwarderPublisher(ABC):
    """
    A forwarder publisher interface.
    """

    @abstractmethod
    def subscribe(self, subscriber: BoboForwarderSubscriber):
        """
        :param subscriber: Subscriber to add to list.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
A phenomenon.
"""

from inspect import signature
from types import MethodType
from typing import Callable, List, Tuple, Optional

from bobocep import BoboError
from bobocep.cep.action.action import BoboAction
from bobocep.cep.phenom.pattern.pattern import BoboPattern

_EXC_NAME_LEN = "name must have a length greater than 0"
_EXC_INVALID_CALL = "datagen callable must have {} parameters, found {}"
_LEN_PARAM_DATAGEN = 2


class BoboPhenomenonError(BoboError):
    """
    A phenomenon error.
    """


class BoboPhenomenon:
    """
    A phenomenon, satisfied by patterns of events, which facilitates the
    generating of complex events
    """

    def __init__(self,
                 name: str,
                 patterns: List[BoboPattern],
                 action: Optional[BoboAction] = None,
                 datagen: Optional[Callable] = None,
                 retain: bool = True):
        """
        :param name: Phenomenon name.
        :param patterns: Phenomenon patterns.
        :param action: Phenomenon action.
        :param datagen: Phenomenon datagen.
        :param retain: If `True`, retains datagen callable as an object
            variable to prevent garbage collection of it.
        """
        super().__init__()

        if len(name) == 0:
            raise BoboPhenomenonError(_EXC_NAME_LEN)

        if datagen is not None:
            len_param_datagen = len(signature(datagen).parameters)

            if len_param_datagen != _LEN_PARAM_DATAGEN:
                raise BoboPhenomenonError(
                    _EXC_INVALID_CALL.format(
                        _LEN_PARAM_DATAGEN,
                        len_param_datagen))

            # Prevent garbage collection of object if callable is a method.
            self._obj = datagen.__self__ \
                if retain and isinstance(datagen, MethodType) else None

        self._name: str = name
        self._patterns: Tuple[BoboPattern, ...] = tuple(patterns)
        self._datagen: Optional[Callable] = datagen
        self._action: Optional[BoboAction] = action
        self._retain: bool = retain

    @property
    def name(self) -> str:
        """
        :return: Phenomenon name.
        """
        return self._name

    @property
    def patterns(self) -> Tuple[BoboPattern, ...]:
        """
        :return: Phenomenon patterns.
        """
        return self._patterns

    @property
    def datagen(self) -> Optional[Callable]:
        """
        :return: Phenomenon datagen, or `None`.
        """
        return self._datagen

    @property
    def action(self) -> Optional[BoboAction]:
        """
        :return: Phenomenon action, or `None`.
        """
        return self._action

    @property
    def retain(self) -> bool:
        """
        :return: True if retains datagen callable; False otherwise.
        """
        return self._retain



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Decider publish-subscribe classes.
"""

from abc import ABC, abstractmethod
from typing import List, Tuple

from bobocep.cep.engine.decider.runserial import BoboRunSerial


class BoboDeciderSubscriber(ABC):
    """
    A decider subscriber interface.
    """

    @abstractmethod
    def on_decider_update(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial],
            local: bool
    ) -> None:
        """
        :param completed: Completed runs.
        :param halted: Halted runs.
        :param updated: Updated runs.
        :param local: `True` if the Decider update occurred locally;
            `False` if the update occurred on a remote (distributed) instance.
        """


class BoboDeciderPublisher(ABC):
    """
    A decider publisher interface.
    """

    @abstractmethod
    def subscribe(self, subscriber: BoboDeciderSubscriber):
        """
        :param subscriber: Subscriber to add to list.
        """

    @abstractmethod
    def snapshot(self) -> Tuple[
        List[BoboRunSerial], List[BoboRunSerial], List[BoboRunSerial]
    ]:
        """
        A snapshot of the current state of the Decider.

        :return: Tuple of cached completed, cached halted, and
            currently partially-completed runs.
            If caching is disabled, the first two lists will be empty.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Action event.
"""

from json import dumps, loads
from typing import Any

from bobocep.cep.event.event import BoboEvent, BoboEventError

_EXC_PRO_LEN = "phenomenon name must have a length greater than 0"
_EXC_PAT_LEN = "pattern name must have a length greater than 0"
_EXC_ACT_LEN = "action name must have a length greater than 0"


class BoboEventAction(BoboEvent):
    """
    An action event.
    """

    TYPE_ACTION = "type_action"

    PHENOMENON_NAME = "phenomenon_name"
    PATTERN_NAME = "pattern_name"
    ACTION_NAME = "action_name"
    SUCCESS = "success"

    def __init__(self,
                 event_id,
                 timestamp: int,
                 data: Any,
                 phenomenon_name: str,
                 pattern_name: str,
                 action_name: str,
                 success: bool):
        """
        :param event_id: The event ID.
        :param timestamp: The event timestamp.
        :param data: The event data.
        :param phenomenon_name: The phenomenon name.
        :param pattern_name: The pattern name.
        :param action_name: The action name.
        :param success: `True` if the action was successful;
            `False` otherwise.

        :raises BoboEventError: If length of phenomenon name is equal to 0.
        :raises BoboEventError: If length of pattern name is equal to 0.
        :raises BoboEventError: If length of action name is equal to 0.
        """
        super().__init__(
            event_id=event_id,
            timestamp=timestamp,
            data=data)

        if len(phenomenon_name) == 0:
            raise BoboEventError(_EXC_PRO_LEN)

        if len(pattern_name) == 0:
            raise BoboEventError(_EXC_PAT_LEN)

        if len(action_name) == 0:
            raise BoboEventError(_EXC_ACT_LEN)

        self._phenomenon_name: str = phenomenon_name
        self._pattern_name: str = pattern_name
        self._action_name: str = action_name
        self._success: bool = success

    def cast(self, dtype: type) -> 'BoboEventAction':
        """
        :param dtype: The type to which the event's data is cast.
        :return: A new BoboEventAction instance with its data cast to `dtype`
            and all other properties identical to the original event.
        """
        return BoboEventAction(
            event_id=self._event_id,
            timestamp=self._timestamp,
            data=dtype(self._data),
            phenomenon_name=self._phenomenon_name,
            pattern_name=self._pattern_name,
            action_name=self._action_name,
            success=self._success
        )

    @property
    def phenomenon_name(self) -> str:
        """
        :return: Phenomenon name.
        """
        return self._phenomenon_name

    @property
    def pattern_name(self) -> str:
        """
        :return: Pattern name.
        """
        return self._pattern_name

    @property
    def action_name(self) -> str:
        """
        :return: Action name.
        """
        return self._action_name

    @property
    def success(self) -> bool:
        """
        :return: `True` if action was executed successfully;
            `False` otherwise.
        """
        return self._success

    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of the event.
        """
        return {
            self.EVENT_TYPE: self.TYPE_ACTION,
            self.EVENT_ID: self.event_id,
            self.TIMESTAMP: self.timestamp,
            self.DATA: self.data,
            self.PHENOMENON_NAME: self.phenomenon_name,
            self.PATTERN_NAME: self.pattern_name,
            self.ACTION_NAME: self.action_name,
            self.SUCCESS: self.success
        }

    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return dumps(self.to_json_dict(), default=lambda o: o.to_json_str())

    @staticmethod
    def from_json_str(j: str) -> 'BoboEventAction':
        """
        :param j: A JSON `str` representation of the event.
        :return: A new instance of the event type.
        """
        return BoboEventAction.from_json_dict(loads(j))

    @staticmethod
    def from_json_dict(d: dict) -> 'BoboEventAction':
        """
        :param d: A JSON `dict` representation of the event.
        :return: A new instance of the event type.
        """
        return BoboEventAction(
            event_id=d[BoboEventAction.EVENT_ID],
            timestamp=d[BoboEventAction.TIMESTAMP],
            data=d[BoboEventAction.DATA],
            phenomenon_name=d[BoboEventAction.PHENOMENON_NAME],
            pattern_name=d[BoboEventAction.PATTERN_NAME],
            action_name=d[BoboEventAction.ACTION_NAME],
            success=d[BoboEventAction.SUCCESS]
        )

    def __str__(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return self.to_json_str()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Producer publish-subscribe classes.
"""

from abc import ABC, abstractmethod

from bobocep.cep.event import BoboEventComplex


class BoboProducerSubscriber(ABC):
    """A producer subscriber interface."""

    @abstractmethod
    def on_producer_update(
            self,
            event: BoboEventComplex,
            local: bool
    ) -> None:
        """
        :param event: Complex event generated by Producer.
        :param local: `True` if the complex event was generated using
            a locally-completed run; `False` otherwise.
        """


class BoboProducerPublisher(ABC):
    """A producer publisher interface."""

    @abstractmethod
    def subscribe(self, subscriber: BoboProducerSubscriber):
        """
        :param subscriber: Subscriber to add to list.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Core classes.
"""

from abc import ABC, abstractmethod


class BoboError(Exception):
    """
    A `BoboCEP` error.
    """


class BoboJSONableError(BoboError):
    """
    A JSONable error.
    """


class BoboJSONable(ABC):
    """
    An abstract interface for JSONable types.
    """

    @abstractmethod
    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of an object of this type.
        """

    @abstractmethod
    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of an object of this type.
        """

    @staticmethod
    @abstractmethod
    def from_json_str(j: str) -> 'BoboJSONable':
        """
        :param j: A JSON `str` representation of an object of this type.
        :return: A new instance of this type.
        """

    @staticmethod
    @abstractmethod
    def from_json_dict(d: dict) -> 'BoboJSONable':
        """
        :param d: A JSON `dict` representation of an object of this type.
        :return: A new instance of this type.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Receiver publish-subscribe classes.
"""

from abc import ABC, abstractmethod

from bobocep.cep.event import BoboEvent


class BoboReceiverSubscriber(ABC):
    """
    A receiver subscriber interface.
    """

    @abstractmethod
    def on_receiver_update(self, event: BoboEvent) -> None:
        """
        :param event: A new BoboEvent instance processed by the receiver.
        """


class BoboReceiverPublisher(ABC):
    """
    A receiver publisher interface.
    """

    @abstractmethod
    def subscribe(self, subscriber: BoboReceiverSubscriber) -> None:
        """
        :param subscriber: Subscriber to add to list.
        """



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name="bobocep",
    version="1.1.0",
    author="r3w0p",
    author_email="rr33ww00pp@gmail.com",
    description="A fault-tolerant Complex Event Processing engine "
                "designed for edge computing in Internet of Things systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/r3w0p/bobocep",
    keywords=[
        "complex event processing",
        "distributed systems",
        "edge computing",
        "fault tolerance",
        "internet of things",
        "web of things"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Monitoring"
    ],
    install_requires=install_requires,
    packages=find_packages(include=["bobocep", "bobocep.*"]),
    test_suite="tests",
    zip_safe=False
)



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Distributed imports.
"""

from bobocep.dist.device import BoboDevice, BoboDeviceError
from bobocep.dist.dist import BoboDistributed, BoboDistributedError, \
    BoboDistributedSystemError, BoboDistributedTimeoutError, \
    BoboDistributedJSONError, BoboDistributedJSONEncodeError, \
    BoboDistributedJSONDecodeError
from bobocep.dist.tcp import BoboDistributedTCP



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Engine task that provides an entry point for data into the system.
"""

from queue import Queue
from threading import RLock
from typing import Optional, Any, List

from bobocep.cep.engine.forwarder.pubsub import BoboForwarderSubscriber
from bobocep.cep.engine.producer.pubsub import BoboProducerSubscriber
from bobocep.cep.engine.receiver.pubsub import BoboReceiverPublisher, \
    BoboReceiverSubscriber
from bobocep.cep.engine.receiver.validator import BoboValidator
from bobocep.cep.engine.task import BoboEngineTaskError, BoboEngineTask
from bobocep.cep.event import BoboEvent, BoboEventSimple, BoboEventComplex, \
    BoboEventAction
from bobocep.cep.gen.event import BoboGenEvent
from bobocep.cep.gen.event_id import BoboGenEventID
from bobocep.cep.gen.timestamp import BoboGenTimestamp

_EXC_QUEUE_FULL = "queue is full (max size: {})"


class BoboReceiverError(BoboEngineTaskError):
    """
    A receiver task error.
    """


class BoboReceiver(BoboEngineTask,
                   BoboReceiverPublisher,
                   BoboProducerSubscriber,
                   BoboForwarderSubscriber):
    """
    A receiver task.
    """

    def __init__(self,
                 validator: BoboValidator,
                 gen_event_id: BoboGenEventID,
                 gen_timestamp: BoboGenTimestamp,
                 gen_event: Optional[BoboGenEvent] = None,
                 max_size: int = 0):
        """
        :param validator: Incoming data validator.
        :param gen_event_id: Event ID generator.
        :param gen_timestamp: Timestamp generator.
        :param gen_event: Event generator (optional).
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        super().__init__()

        self._lock: RLock = RLock()
        self._closed: bool = False
        self._subscribers: List[BoboReceiverSubscriber] = []

        self._validator: BoboValidator = validator
        self._gen_event_id: BoboGenEventID = gen_event_id
        self._gen_timestamp: BoboGenTimestamp = gen_timestamp
        self._gen_event: Optional[BoboGenEvent] = gen_event

        self._max_size: int = max(0, max_size)
        self._queue: Queue[Any] = Queue(self._max_size)

    def subscribe(self, subscriber: BoboReceiverSubscriber) -> None:
        """
        :param subscriber: Subscriber to Receiver data.
        """
        with self._lock:
            if subscriber not in self._subscribers:
                self._subscribers.append(subscriber)

    def add_data(self, data: Any) -> None:
        """
        :param data: Data to add to the receiver.

        :raises BoboReceiverError: If receiver queue is full.
        """
        with self._lock:
            if self._closed:
                return

            if not self._queue.full():
                self._queue.put(data)
            else:
                raise BoboReceiverError(
                    _EXC_QUEUE_FULL.format(self._max_size))

    def _process_data(self, data: Any) -> None:
        """
        :param data: Data to process.
        """
        if not self._validator.is_valid(data):
            return None

        if isinstance(data, BoboEvent):
            event = data
        else:
            event = BoboEventSimple(
                event_id=self._gen_event_id.generate(),
                timestamp=self._gen_timestamp.generate(),
                data=data)

        for subscriber in self._subscribers:
            subscriber.on_receiver_update(event)

    def update(self) -> bool:
        """
        Processes data from its queue, if any.
        Also processes a generated event if receiver is set to generate any.
        A BoboEventSimple instance is produced for data if they pass validation
        and are not already BoboEvent instances. Valid data are then sent to
        receiver subscribers.

        :return: `True` if queue or generated data are processed;
            `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return False

            data: Any = None
            event_gen: Optional[BoboEvent] = None

            if not self._queue.empty():
                data = self._queue.get_nowait()
                self._process_data(data)

            if self._gen_event is not None:
                event_gen = self._gen_event.maybe_generate(
                    self._gen_event_id.generate())

                if event_gen is not None:
                    self._process_data(event_gen)

            return data is not None or event_gen is not None

    def size(self) -> int:
        """
        :return: Queue size.
        """
        with self._lock:
            return self._queue.qsize()

    def close(self) -> None:
        """
        Closes the Receiver.
        """
        with self._lock:
            self._closed = True

    def is_closed(self) -> bool:
        """
        :return: `True` if Receiver is closed; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def on_producer_update(
            self,
            event: BoboEventComplex,
            local: bool
    ) -> None:
        """
        :param event: Complex event generated by Producer.
        :param local: `True` if the complex event was generated using
            a locally-completed run; `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return

            self.add_data(event)

    def on_forwarder_update(self, event: BoboEventAction) -> None:
        """
        :param event: Action event generated by Forwarder.
        """
        with self._lock:
            if self._closed:
                return

            self.add_data(event)



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Receiver data validators.
"""

from abc import ABC, abstractmethod
from json import dumps
from typing import Any, List, Tuple

from jsonschema import validate as jsonschema_validate  # type: ignore
from jsonschema.exceptions import ValidationError, SchemaError  # type: ignore

from bobocep import BoboError
from bobocep.cep.event import BoboEvent


class BoboValidatorError(BoboError):
    """
    A validator error.
    """


class BoboValidator(ABC):
    """An abstract validator."""

    @abstractmethod
    def is_valid(self, data: Any) -> bool:
        """
        :param data: Data to validate.
        :return: `True` if data are valid; `False` otherwise.
        """


class BoboValidatorAll(BoboValidator):
    """
    Validator that accepts all data.
    """

    def is_valid(self, data: Any) -> bool:
        """
        :return: Always returns `True`.
        """
        return True


class BoboValidatorJSONable(BoboValidator):
    """
    Validates whether the data type is JSONable. If the data are a
    BoboEvent, then the event's data are checked instead.
    """

    def is_valid(self, data: Any) -> bool:
        """
        :return: `True` if data are valid JSON; `False` otherwise.
        """
        if isinstance(data, BoboEvent):
            data = data.data

        try:
            dumps(data)

        except (RecursionError, TypeError, ValueError):
            return False

        return True


class BoboValidatorType(BoboValidator):
    """
    Validates whether the entity type matches any of the given data types.
    If the data are a BoboEvent, then the event's data are checked instead.
    """

    def __init__(self,
                 types: List[type],
                 subtype: bool = True):
        """
        :param types: The types of which the data must match at least one
            for the data to be valid.
        :param subtype: If `True`, subtypes of a type are also valid.
            If `False`, data must match exact type.
        """
        super().__init__()

        self._types: Tuple[type, ...] = tuple(types)
        self._subtype = subtype

    def is_valid(self, data: Any) -> bool:
        """
        :return: `True` if the data matches a type; `False` otherwise.
        """
        if isinstance(data, BoboEvent):
            data = data.data

        if self._subtype:
            return any(isinstance(data, t) for t in self._types)
        else:
            return any(type(data) == t for t in self._types)


class BoboValidatorJSONSchema(BoboValidatorJSONable):
    """
    Validates whether the data type is valid with respect to
    a given JSON Schema. If the data are a BoboEvent,
    then the event's data are checked instead.
    """

    def __init__(self, schema: dict):
        """
        :param schema: The JSON schema against which to compare data.
        """
        super().__init__()

        self._schema: dict = schema

    def is_valid(self, data: Any) -> bool:
        """
        :return: `True` if data are valid as per the JSON schema;
                 `False` otherwise.

        :raises: BoboValidatorError: Invalid JSON schema.
        """
        try:
            jsonschema_validate(instance=data, schema=self._schema)

        except ValidationError:
            return False

        except SchemaError as e:
            raise BoboValidatorError(e)

        return True



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Top-level package for BoboCEP.
"""

__author__ = """r3w0p"""
__email__ = "rr33ww00pp@gmail.com"
__version__ = "1.1.0"

from bobocep.bobocep import BoboError, BoboJSONable, BoboJSONableError



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Handlers that coordinate the execution of actions.
"""

import logging
from abc import ABC, abstractmethod
from queue import Queue
from threading import RLock
from typing import Any, Optional, Tuple, NamedTuple

from bobocep import BoboError
from bobocep.cep.action.action import BoboAction
from bobocep.cep.event import BoboEventComplex

_EXC_QUEUE_FULL = "queue is full (max size: {})"


class BoboActionHandlerError(BoboError):
    """
    An action handler error.
    """


class BoboHandlerResponse(NamedTuple):
    """
    A handler response to action execution.
    """
    action_name: str
    complex_event: BoboEventComplex
    success: bool
    data: Any


class BoboActionHandler(ABC):
    """
    An abstract action handler.
    """

    def __init__(self, max_size: int = 0):
        """
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        super().__init__()
        self._closed: bool = False
        self._lock: RLock = RLock()
        self._max_size = max(0, max_size)

    def handle(self,
               action: BoboAction,
               event: BoboEventComplex) -> Any:
        """
        Handle an action.

        :param action: The action to handle.
        :param event: The complex event that caused the action to trigger.

        :return: A return value from handling the action.
        """
        with self._lock:
            return self._execute_action(action, event)

    def is_closed(self) -> bool:
        """
        :return: `True` if handler is closed; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def close(self) -> None:
        """
        Close the handler.
        """
        with self._lock:
            try:
                self._on_closing()
            finally:
                self._closed = True

    @abstractmethod
    def _on_closing(self) -> None:
        """
        Close the handler.
        """

    @abstractmethod
    def _execute_action(
            self,
            action: BoboAction,
            event: BoboEventComplex) -> Any:
        """
        Execute an action.

        :param action: The action to execute.
        :param event: The complex event that caused the action to trigger.

        :return: A return value from executing the action.
        """

    @abstractmethod
    def _get_queue(self) -> Queue:
        """
        :return: The handler queue.
        """

    def get_handler_response(self) -> Optional[BoboHandlerResponse]:
        """
        :return: Action response from queue, or `None` if queue is empty.
        """
        with self._lock:
            queue = self._get_queue()
            if not queue.empty():
                return queue.get_nowait()
            return None

    def size(self) -> int:
        """
        :return: The size of the handler queue.
        """
        with self._lock:
            return self._get_queue().qsize()


class BoboActionHandlerBlocking(BoboActionHandler):
    """
    An action handler that blocks during action execution.
    """

    def __init__(self, max_size: int = 0):
        """
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        super().__init__(max_size)

        self._queue: "Queue[BoboHandlerResponse]" = Queue(self._max_size)

    def _execute_action(self,
                        action: BoboAction,
                        event: BoboEventComplex) -> None:
        """
        :param action: Action to execute.
        :param event: Complex event associated with the action.
        """
        action_ret: Tuple[bool, Any] = action.execute(event)

        hres = BoboHandlerResponse(
            action_name=action.name,
            complex_event=event,
            success=action_ret[0],
            data=action_ret[1]
        )

        if not self._queue.full():
            self._queue.put(hres)
        else:
            raise BoboActionHandlerError(
                _EXC_QUEUE_FULL.format(self._max_size))

    def _on_closing(self) -> None:
        """
        Action on closing the handler.
        """
        pass

    def _get_queue(self) -> Queue:
        """
        :return: Handler queue.
        """
        return self._queue


def _pool_execute_action(
        queue: Queue,
        action: BoboAction,
        event: BoboEventComplex,
        max_size: int
) -> None:
    """
    :param queue: A queue in which to put the action response that is
        returned after execution.
    :param action: The action to execute.
    :param event: The complex event that triggered the action being executed.
    :param max_size: Maximum queue size.
    """
    try:
        action_ret: Tuple[bool, Any] = action.execute(event)

    except (Exception,) as e:
        logging.error(e)
        raise e

    hres = BoboHandlerResponse(
        action_name=action.name,
        complex_event=event,
        success=action_ret[0],
        data=action_ret[1]
    )

    if not queue.full():
        queue.put(hres)
    else:
        raise BoboActionHandlerError(
            _EXC_QUEUE_FULL.format(max_size))


class BoboActionHandlerMultithreading(BoboActionHandler):
    """
    An action handler that uses multithreading for action execution.
    """
    from multiprocessing.pool import AsyncResult

    def __init__(self, threads: int, max_size: int = 0):
        """
        :param threads: Number of thread processes to use
            for handling actions.
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        from multiprocessing.pool import ThreadPool
        super().__init__(max_size)

        self._threads = threads
        self._pool: ThreadPool = ThreadPool(processes=threads)
        self._queue: Queue[BoboHandlerResponse] = Queue()

    def join(self) -> None:
        """
        Join the multiprocessing pool.
        """
        with self._lock:
            self._pool.join()

    def _execute_action(self,
                        action: BoboAction,
                        event: BoboEventComplex) -> AsyncResult:
        """
        :param action: Action to execute.
        :param event: Complex event associated with the action.

        :return: Result from asynchronous action execution.
        """
        # The queue size is checked manually because queue.full() does not
        # seem to work properly...
        if self._max_size > 0 and (self._queue.qsize() >= self._max_size):
            raise BoboActionHandlerError(
                _EXC_QUEUE_FULL.format(self._max_size))

        return self._pool.starmap_async(
            _pool_execute_action, [
                (self._queue, action, event, self._max_size)
            ])

    def _get_queue(self) -> Queue:
        """
        :return: Handler queue.
        """
        return self._queue

    def _on_closing(self) -> None:
        """
        Action on closing the handler.
        """
        self._pool.close()


class BoboActionHandlerMultiprocessing(BoboActionHandler):
    """
    An action handler that uses multiprocessing for action execution.
    """
    from multiprocessing.pool import AsyncResult

    def __init__(self, processes: int, max_size: int = 0):
        """
        :param processes: Number of multicore processes to use
            for handling actions.
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        from multiprocessing import Manager, Pool
        super().__init__(max_size)

        self._processes = processes
        self._pool = Pool(processes=processes)
        self._manager = Manager()
        self._queue: "Queue[BoboHandlerResponse]" = self._manager.Queue()

    def join(self) -> None:
        """
        Join the multiprocessing pool.
        """
        with self._lock:
            self._pool.join()

    def _execute_action(self,
                        action: BoboAction,
                        event: BoboEventComplex) -> AsyncResult:
        """
        :param action: Action to execute.
        :param event: Complex event associated with the action.

        :return: Result from asynchronous action execution.
        """
        # The queue size is checked manually before action execution because
        # a 'queue full' error within a running process would not be visible
        # outside of the process itself. Also, a 'maxsize' parameter for the
        # Manager queue does not appear to work correctly e.g. a max_size of 1
        # causes unit tests to 'hang'.
        if self._max_size > 0 and (self._queue.qsize() >= self._max_size):
            raise BoboActionHandlerError(
                _EXC_QUEUE_FULL.format(self._max_size))

        return self._pool.starmap_async(
            _pool_execute_action, [
                (self._queue, action, event, self._max_size)
            ])

    def _get_queue(self) -> Queue:
        """
        :return: Handler queue.
        """
        return self._queue

    def _on_closing(self) -> None:
        """
        Action on closing the handler.
        """
        self._pool.close()
